# coding: utf-8
import os
import time
import hashlib


class BaseGenerator(object):

    def __init__(self, local=False):
        self.local = local


class RandomGenerator(BaseGenerator):

    def __call__(self):
        md5 = hashlib.md5(os.urandom(32).encode('hex')).hexdigest()
        if self.local:
            return '%s/%s/%s' % (md5[:2], md5[2:4], md5[4:])
        return md5


class DatetimeGenerator(BaseGenerator):
    """ 多进程下会有Bug """

    def __init__(self, local=False):
        self.local = local
        self.tpl = '%Y%m%d/%H%M%S' if self.local else '%Y%m%d%H%M%S'

    stamp = long(time.time())
    index = 0

    @staticmethod
    def get_index():
        if long(time.time()) > DatetimeGenerator.stamp:
            DatetimeGenerator.stamp = time.time()
            DatetimeGenerator.index = 0
        DatetimeGenerator.index += 1
        return DatetimeGenerator.index

    def __call__(self):
        return '%s%02d' % (
            time.strftime(self.tpl), DatetimeGenerator.get_index())



# coding: utf-8
import base64
import hashlib
from flask import current_app
from mongoengine import signals
from mongoengine.fields import ListField, EmbeddedDocument, StringField
from mongoengine.base.fields import BaseField
from werkzeug.datastructures import FileStorage
from .generators import BaseGenerator, RandomGenerator
from .storages import get_storage

__all__ = [
    'XFileField', 'XImageField', 'Base64ImageField', 'XListField',
    'AreaField', 'set_filename_generator', 'FileProxy', 'ImageProxy',
    'Base64ImageProxy',
]
DEFAULT_ALLOWS = ['txt', 'bz2', 'gz', 'tar', 'zip', 'rar', 'apk',
                  'jpg', 'jpeg', 'png', 'gif', 'bmp']
DEFAULT_IMAGE_ALLOWS = ['jpg', 'jpeg', 'png', 'gif', 'bmp']


def is_empty(fd):
    fd.seek(0)
    first_char = fd.read(1)
    fd.seek(0)
    return not bool(first_char)


_filename_generators = {}


def set_filename_generator(key, generator):
    global _filename_generators
    _filename_generators[key] = generator


def get_filename_generator(key, local, generator=None):
    global _filename_generators
    if not generator:
        generator = RandomGenerator
    if key not in _filename_generators:
        _filename_generators[key] = generator(local)
    return _filename_generators[key]


class FileProxy(object):

    def __init__(self, instance, value=None):
        self.instance = instance
        self.filename = ''
        self.process(value)

    def __nonzero__(self):
        return True if self.filename else False

    @property
    def path(self):
        return self.instance.get_path(self.filename)

    @property
    def src(self):
        return self.instance.get_link(self.filename, source=True)

    @property
    def link(self):
        return self.instance.get_link(self.filename, source=True)

    def get_link(self, width=0, height=0, ystart=0, yend=0):
        return self.instance.get_link(self.filename, source=True)

    @property
    def content(self):
        return self.instance.get_content(self.filename)

    @property
    def md5(self):
        content = self.content
        if content:
            return hashlib.md5(content).hexdigest()
        return ''

    def process(self, value=None):
        if isinstance(value, FileStorage):
            self._process(stream=value.stream,
                          format=value.filename.split('.')[-1])
        elif isinstance(value, dict):
            self._process(stream=value.get('stream'),
                          format=value.get('format'))
        elif isinstance(value, (tuple, list)) and len(value) == 2:
            self._process(stream=value[0], format=value[1])
        elif isinstance(value, (str, unicode)):
            self._process(filename=value)
        elif hasattr(value, 'read'):
            self._process(stream=value)
        elif not value:
            self._process()
        elif isinstance(value, FileProxy):
            self._process(filename=value.filename)
        else:
            raise ValueError('Can not support type(%s)' % str(value))

    def _process(self, stream=None, format=None, filename=None):
        if stream is not None:
            if not is_empty(stream):
                self.remove()
                if self.instance.rename or not self.filename:
                    self.filename = self.instance.put(stream, format=format)
                else:
                    self.instance.put(stream, filename=self.filename)
        elif filename is not None:
            self.remove()
            self.filename = filename
        else:
            self.remove()
            self.filename = ''

    def remove(self):
        self.instance.remove(self.filename)

    def __unicode__(self):
        return self.filename


class XFileField(BaseField):

    proxy_class = FileProxy
    default_allows = DEFAULT_ALLOWS

    def __init__(self, max_size=2 * 1024 * 1024, auto_remove=None,
                 rename=True, allows=None, config='UPLOADS',
                 filename_generator=None, place='', **kwargs):
        self.max_size = max_size
        self.auto_remove = auto_remove
        self.rename = rename
        self.allows = allows or self.default_allows
        self.config = config
        self._filename_generator = filename_generator
        self.place = place
        super(XFileField, self).__init__(**kwargs)

    @property
    def storage(self):
        if not hasattr(self, '_storage'):
            config = current_app.config.get(self.config)
            if not config:
                raise ValueError('Please set %s in config.' % self.config)
            self._storage = get_storage(self.config, config)
        return self._storage

    @property
    def filename_generator(self):
        if not self._filename_generator or isinstance(
                self._filename_generator, type):
            config = current_app.config.get(self.config)
            if not config:
                raise ValueError('Please set %s in config.' % self.config)
            self._filename_generator = get_filename_generator(
                self.config, config['type'] == 'local',
                generator=self._filename_generator)
        return self._filename_generator

    @property
    def is_auto_remove(self):
        if self.auto_remove is not None:
            return self.auto_remove
        return current_app.config.get(self.config).get('auto_remove', True)

    def get_path(self, filename):
        if filename:
            return self.storage.get_path(filename)
        return ''

    def get_link(self, filename, **kwargs):
        if filename:
            if filename.startswith('http://'):
                return filename
            return self.storage.get_link(filename, **kwargs)
        return ''

    def get_content(self, filename):
        if filename:
            return self.storage.get(filename)

    def put(self, stream, format=None, filename=None):
        if not filename:
            filename = self.filename_generator()
            if format:
                filename = '%s.%s' % (filename, format)
        filename = self.storage.put(filename, stream.read())
        return filename

    def remove(self, filename):
        if filename and self.is_auto_remove:
            self.storage.remove(filename)

    def register_signals(self, instance):
        if not hasattr(self, '_instance') and instance is not None:
            self._instance = instance
            signals.pre_delete.connect(
                self.pre_delete, sender=self._instance.__class__)

    def pre_delete(self, sender, document, **kwargs):
        obj = document._data.get(self.name)
        if isinstance(obj, self.proxy_class) and self.is_auto_remove:
            obj.remove()

    def _get(self, instance):
        key = self.name
        obj = instance._data.get(key)
        if not isinstance(obj, self.proxy_class) or obj is None:
            obj = self.proxy_class(self, obj)
            instance._data[key] = obj
        return instance._data[key]

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self._get(instance)

    def __set__(self, instance, value):
        self.register_signals(instance)

        key = self.name
        obj = instance._data.get(key)
        if isinstance(value, self.proxy_class) and value.instance == self:
            if obj and id(obj) != value and str(value) != str(value):
                self._get(instance).remove()
            instance._data[key] = value
        elif not isinstance(obj, self.proxy_class):
            obj = self.proxy_class(self, obj)
            obj.process(value)
            instance._data[key] = obj
        else:
            obj.process(value)
        instance._mark_as_changed(key)

    def to_mongo(self, value):
        if isinstance(value, self.proxy_class):
            return value.filename
        return value

    def to_python(self, value):
        if not isinstance(value, self.proxy_class):
            return self.proxy_class(self, value)
        return value


class ImageProxy(FileProxy):

    @property
    def link(self):
        return self.instance.get_link(self.filename)

    def get_link(self, width=0, height=0, ystart=0, yend=0):
        return self.instance.get_link(
            self.filename, width=width, height=height,
            ystart=ystart, yend=yend)


class XImageField(XFileField):
    proxy_class = ImageProxy
    default_allows = DEFAULT_IMAGE_ALLOWS


class Base64ImageProxy(FileProxy):

    @property
    def link(self):
        return self.filename

    @property
    def base64(self):
        return ('base://' + self.filename) if self.filename else ''

    def get_link(self, width=0, height=0, ystart=0, yend=0):
        return self.filename


class Base64ImageField(XImageField):
    """ Base64图片字段 """

    proxy_class = Base64ImageProxy

    def get_path(self, filename):
        return filename

    def get_link(self, filename, **kwargs):
        return filename

    def get_content(self, filename):
        if filename and filename.find('base64,') != -1:
            return filename[filename.index('base64,'):].decode('base64')

    def put(self, stream, format='png', filename=None):
        filename = 'data:image/%s;base64,' % format
        filename += base64.b64encode(stream.read())
        return filename

    def remove(self, filename):
        pass

    def to_mongo(self, value):
        if isinstance(value, self.proxy_class):
            return value.filename
        return value

    def to_python(self, value):
        if not isinstance(value, self.proxy_class):
            return self.proxy_class(self, value)
        return value


class XListField(ListField):

    def register_signals(self, instance):
        if not hasattr(self, '_instance') and instance is not None:
            self._instance = instance
            signals.pre_delete.connect(
                self.pre_delete, sender=self._instance.__class__)

    def pre_delete(self, sender, document, **kwargs):
        values = document._data.get(self.name) or []
        for obj in values:
            if isinstance(obj, EmbeddedDocument):
                for field in obj._fields:
                    attr = getattr(type(obj), field)
                    value = getattr(obj, field)
                    if hasattr(value, 'remove') \
                            and isinstance(value, attr.proxy_class) \
                            and attr.is_auto_remove:
                        value.remove()
            elif isinstance(self.field, XFileField) \
                    and isinstance(obj, self.field.proxy_class) \
                    and self.field.is_auto_remove:
                obj.remove()

    def __set__(self, instance, value):
        self.register_signals(instance)

        value = [self.field.to_python(x) for x in value] if value else []
        data = getattr(instance, self.name)
        if data:
            for item in data:
                if item not in value:
                    if isinstance(item, EmbeddedDocument):
                        for field in item._fields:
                            attr = getattr(item, field)
                            if hasattr(attr, 'remove'):
                                attr.remove()
                    elif hasattr(item, 'remove'):
                        item.remove()
        super(XListField, self).__set__(instance, value)


class AreaField(StringField):
    pass



# coding: utf-8
from flask.ext import mongoengine
from . import fields, pagination


def _include_custom(obj):
    for key in fields.__all__:
        if not hasattr(obj, key):
            setattr(obj, key, getattr(fields, key))
    setattr(obj, 'abstract', abstract)


class Chocies(object):

    def __init__(self, **kwargs):
        self.CHOICES = []
        for key, value in kwargs.iteritems():
            self.CHOICES.append((key, value))
            setattr(self, key.upper(), key)
        self.DICT = dict(self.CHOICES)
        self.VALUES = self.DICT.keys()

    def text(self, key):
        return self.DICT.get(key)


class MongoEngine(mongoengine.MongoEngine):

    def __init__(self, app=None):
        super(MongoEngine, self).__init__(app)
        _include_custom(self)

        self.Document = Document
        self.DynamicDocument = DynamicDocument

    def choices(self, **kwargs):
        return Chocies(**kwargs)


class BaseQuerySet(mongoengine.BaseQuerySet):

    def paginate(self, **kwargs):
        return pagination.Pagination(self, **kwargs)


class Document(mongoengine.Document):

    meta = {'abstract': True,
            'queryset_class': BaseQuerySet}


class DynamicDocument(mongoengine.DynamicDocument):

    meta = {'abstract': True,
            'queryset_class': BaseQuerySet}


def abstract(model):
    model._meta['abstract'] = True



# coding: utf-8
import os
import shutil
from flask import current_app


def is_empty_folder(folder):
    for root, dirs, files in os.walk(folder):
        if len(dirs) == 0 and len(files) == 0:
            return True
        return False


def load_file(name):
    if os.path.exists(name):
        with open(name) as fd:
            return fd.read()


def save_file(name, content):
    dirname = os.path.dirname(name)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(name, 'w+') as fd:
        fd.write(content)


def remove_file(name, path):
    if os.path.exists(name):
        os.remove(name)

        folder = os.path.dirname(name)
        while folder != path:
            if is_empty_folder(folder):
                shutil.rmtree(folder)
            folder = os.path.dirname(folder)


class BaseFile(object):

    def __init__(self, conf):
        self.conf = conf

    def get_path(self, name):
        return name

    def get_link(self, name, **kwargs):
        return self.conf['link'] % name


class LocalFile(BaseFile):

    def __init__(self, conf):
        super(LocalFile, self).__init__(conf)
        self.path = self.conf['path']

    def get_path(self, name):
        return os.path.join(self.path, name)

    def get(self, name):
        return load_file(self.get_path(name))

    def put(self, name, content):
        save_file(self.get_path(name), content)
        return name

    def remove(self, name):
        return remove_file(self.get_path(name), self.conf['path'])


class OSSFile(BaseFile):

    def __init__(self, conf):
        super(OSSFile, self).__init__(conf)
        from oss.oss_api import OssAPI
        self.oss = OssAPI(
            conf['host'],
            conf['access_id'],
            conf['secret_access_key'],
        )

    def get_link(self, name, width=0, height=0, ystart=0, yend=0,
                 source=False):
        link = self.conf['link'] % name
        if source:
            return link

        format = name.split('.')[-1]
        if format not in ['jpg', 'jpeg', 'png', 'gif', 'bmp']:
            format = 'jpg'

        attrs = []
        if width != 0:
            attrs.append('w_%d' % width)
        if height != 0:
            attrs.append('h_%d' % height)
        if attrs:
            if width and height:
                attrs.append('m_fill')
            attrs.append('limit_0')
        if attrs:
            return link + '?x-oss-process=image/resize,' + \
                ','.join(attrs) + '/format,' + format
        return link

    def get(self, name):
        res = self.oss.get_object(self.conf['bucket'], name)
        if (res.status / 100) == 2:
            return res.read()
        return ''

    def put(self, name, content, type=None):
        name = '%s.%s' % (name, type) if type is not None else name
        if not name.startswith(self.conf.get('prefix')):
            name = self.conf.get('prefix', '') + name
        res = self.oss.put_object_from_string(
            self.conf['bucket'], name, content)
        if (res.status / 100) == 2:
            return name
        current_app.logger.error('oss error: \n' + res.read())
        return ''

    def remove(self, name):
        res = self.oss.delete_object(self.conf['bucket'], name)
        return (res.status / 100) == 2


def _get_storage(conf):
    if conf['type'] == 'local':
        return LocalFile(conf)
    elif conf['type'] == 'oss':
        return OSSFile(conf)
    raise ValueError('Cannot support type (%s) of config' % conf['type'])


_storages = {}


def get_storage(key, conf):
    global _storages
    if key not in _storages:
        _storages[key] = _get_storage(conf)
    return _storages[key]



# coding: utf-8
from chiki_base import is_ajax, json_success
from flask import request, url_for, render_template
from flask.ext.mongoengine.pagination import Pagination as _Pagination


def success(_data=None, **kwargs):
    if kwargs.get('__external'):
        kwargs.setdefault('code', 0)
        kwargs.setdefault('key', 'SUCCESS')
        return dict(**kwargs)
    res = dict(code=0, key='SUCCESS')
    if _data is not None:
        res['data'] = _data
    elif kwargs:
        res['data'] = kwargs
    return res


class Pagination(_Pagination):

    def __init__(self, iterable, page=None, per_page=None,
                 endpoint=None, **kwargs):
        page = page or max(1, request.args.get('page', 1, int))
        per_page = per_page or max(
            1, min(100, request.args.get('per_page', 10, int)))
        super(Pagination, self).__init__(iterable, page, per_page)
        self.endpoint = endpoint if endpoint else request.endpoint
        self.kwargs = kwargs

    @property
    def has_pages(self):
        return self.pages > 1

    @property
    def next_link(self):
        if self.has_next:
            return self.get_link(self.next_num)
        return ''

    def get_link(self, page):
        return url_for(self.endpoint, page=page,
                       per_page=self.per_page, **self.kwargs)

    def iter_links(self, pages=5, next=False, last=True,
                   first_page='首页', prev_page='上页',
                   last_page='尾页', next_page='下页'):
        if last:
            yield first_page, self.get_link(1) if self.page > 1 else None
        if next:
            yield prev_page, self.get_link(
                self.prev_num) if self.has_prev else None

        start = max(1, self.page - (pages - 1) / 2)
        end = min(self.pages + 1, start + pages)
        start = max(1, end - pages)
        for i in range(start, end):
            yield i, self.get_link(i)

        if next:
            yield next_page, self.get_link(
                self.next_num) if self.has_next else None
        if last:
            yield last_page, self.get_link(
                self.pages) if self.page < self.pages else None

    def render(self, tpl, **kwargs):
        kwargs.update(self.kwargs)
        if is_ajax():
            return json_success(
                html=render_template(tpl, pag=self, **kwargs),
                next=self.next_link,
            )
        return render_template(tpl, pag=self, **kwargs)

    def json(self, tojson=lambda x: x.json, **kwargs):
        return success(
            items=[tojson(x) for x in self.items],
            next=self.next_link, **kwargs)
